"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
var error_1 = require("container.ts/lib/error");
var validate_1 = require("container.ts/lib/validate");
var debug = require("debug");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var pdu = require("../pdu");
/** Master error codes. */
var EMasterError;
(function (EMasterError) {
    EMasterError[EMasterError["Timeout"] = 0] = "Timeout";
    EMasterError[EMasterError["Write"] = 1] = "Write";
})(EMasterError = exports.EMasterError || (exports.EMasterError = {}));
/** Modbus ADU error class. */
var MasterError = /** @class */ (function (_super) {
    __extends(MasterError, _super);
    function MasterError(code, cause, context) {
        var _this = _super.call(this, { name: "MasterError", value: __assign({ code: code }, context) }, cause) || this;
        _this.isMasterError = true;
        return _this;
    }
    return MasterError;
}(error_1.ErrorChain));
exports.MasterError = MasterError;
/** Master log names. */
var EMasterLog;
(function (EMasterLog) {
    EMasterLog["Request"] = "ModbusMasterRequest";
    EMasterLog["Response"] = "ModbusMasterResponse";
    EMasterLog["Exception"] = "ModbusMasterException";
    EMasterLog["Error"] = "MasterMasterError";
})(EMasterLog = exports.EMasterLog || (exports.EMasterLog = {}));
/** Master log interface. */
var MasterLog = /** @class */ (function () {
    function MasterLog() {
        this.debug = debug("modbus.ts");
    }
    MasterLog.prototype.request = function (request, errorCount) {
        this.debug(EMasterLog.Request, request, errorCount);
    };
    MasterLog.prototype.response = function (response) {
        this.debug(EMasterLog.Response, response);
    };
    MasterLog.prototype.exception = function (exception) {
        this.debug(EMasterLog.Exception, exception);
    };
    MasterLog.prototype.error = function (error) {
        this.debug(EMasterLog.Error, error);
    };
    MasterLog.prototype.bytesTransmitted = function (value) { };
    MasterLog.prototype.bytesReceived = function (value) { };
    return MasterLog;
}());
exports.MasterLog = MasterLog;
/** Modbus abstract ADU master. */
var Master = /** @class */ (function () {
    function Master(options, logConstructor) {
        if (logConstructor === void 0) { logConstructor = MasterLog; }
        /** Responses/exceptions stream. */
        this.receive = new rxjs_1.Subject();
        /** Requests stream. */
        this.transmit = new rxjs_1.Subject();
        /** Internal transmit input queue. */
        this.queueIn = new rxjs_1.Subject();
        /** Internal transmit output queue. */
        this.queueOut = new rxjs_1.Subject();
        /** Queue ID counter. */
        this.queueCounter = 0;
        /** Internal buffer. */
        this.buffer = Buffer.allocUnsafe(0);
        this.retry = options.retry != null ? this.isRetry(options.retry) : 0;
        this.timeout = options.timeout != null ? this.isTimeout(options.timeout) : 5000;
        this.retryWhen = this.isRetryWhen(options.retryWhen || this.masterDefaultRetryWhen.bind(this));
        this.log = options.log || new logConstructor();
        this.masterReset();
    }
    /** Returns true if error is RxJS Timeout. */
    Master.isTimeoutError = function (error) {
        var instanceOf = error instanceof rxjs_1.TimeoutError;
        var hasName = error.name === "TimeoutError";
        return instanceOf || hasName;
    };
    /** Returns true if error is MasterError. */
    Master.isMasterError = function (error) {
        var instanceOf = error instanceof MasterError;
        var hasProperty = !!error.isMasterError;
        return instanceOf || hasProperty;
    };
    /**
     * This function code is used to read from 1 to 2000 contiguous status of coils in a remote device.
     * @param startingAddress Starting address.
     * @param quantityOfCoils Quantity of coils.
     * @param options Request options.
     */
    Master.prototype.readCoils = function (startingAddress, quantityOfCoils, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.readCoils(startingAddress, quantityOfCoils);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * This function code is used to read from 1 to 2000 contiguous status of discrete inputs in a remote device.
     * @param startingAddress Starting address.
     * @param quantityOfInputs Quantity of inputs.
     * @param options Request options.
     */
    Master.prototype.readDiscreteInputs = function (startingAddress, quantityOfInputs, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.readDiscreteInputs(startingAddress, quantityOfInputs);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * This function code is used to read  the contents of a contiguous block of holding registers in a remote device.
     * @param startingAddress Starting address.
     * @param quantityOfRegisters Quantity of registers.
     * @param options Request options.
     */
    Master.prototype.readHoldingRegisters = function (startingAddress, quantityOfRegisters, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.readHoldingRegisters(startingAddress, quantityOfRegisters);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * This function code is used to read from 1 to 125 contiguous input registers in a remote device.
     * @param startingAddress Starting address.
     * @param quantityOfRegisters Quantity of registers.
     * @param options Request options.
     */
    Master.prototype.readInputRegisters = function (startingAddress, quantityOfRegisters, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.readInputRegisters(startingAddress, quantityOfRegisters);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * This function code is used to write a single output to either ON or OFF in a remote device.
     * @param outputAddress Output address.
     * @param outputValue  Output value.
     * @param options Request options.
     */
    Master.prototype.writeSingleCoil = function (outputAddress, outputValue, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.writeSingleCoil(outputAddress, outputValue);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * This function code is used to write a single holding register in a remote device.
     * @param registerAddress Register address.
     * @param registerValue Register value.
     * @param options Request options.
     */
    Master.prototype.writeSingleRegister = function (registerAddress, registerValue, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.writeSingleRegister(registerAddress, registerValue);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * This function code is used to force each coil in a sequence of coils to either ON or OFF in a remote device.
     * @param startingAddress Starting address.
     * @param outputValues Output values.
     * @param options Request options.
     */
    Master.prototype.writeMultipleCoils = function (startingAddress, outputValues, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.writeMultipleCoils(startingAddress, outputValues);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * This function code is used to write a block of contiguous registers (1 to 123 registers) in a remote device.
     * @param startingAddress Starting address.
     * @param registerValues Register values.
     * @param options Request options.
     */
    Master.prototype.writeMultipleRegisters = function (startingAddress, registerValues, options) {
        if (options === void 0) { options = {}; }
        var pduRequest = pdu.Master.writeMultipleRegisters(startingAddress, registerValues);
        var request = this.masterSetupRequest(pduRequest.functionCode, pduRequest.buffer);
        return this.masterOnQueue(request, options);
    };
    /**
     * Default retryWhen callback for conditional retries.
     * This will allow retries in cases of timeout and throw all other errors.
     */
    Master.prototype.masterDefaultRetryWhen = function (master, retry, errorCount, error, request) {
        if (Master.isTimeoutError(error)) {
            // If error is a timeout, retry up to limit.
            if (errorCount >= retry) {
                throw new MasterError(EMasterError.Timeout, error, { request: request });
            }
        }
        else if (Master.isMasterError(error)) {
            // If error is a master error, rethrow now.
            throw error;
        }
        else {
            // Wrap and rethrow unknown errors.
            throw new MasterError(error.code, error, { request: request });
        }
    };
    /** Reset master internal state, must be called by subclass when (re)creating a connection. */
    Master.prototype.masterReset = function () {
        var _this = this;
        this.masterDestroy();
        this.receive = new rxjs_1.Subject();
        this.transmit = new rxjs_1.Subject();
        this.queueIn = new rxjs_1.Subject();
        this.queueOut = new rxjs_1.Subject();
        // Subscribe to queue input, make requests in sequence.
        this.queueIn
            .pipe(operators_1.concatMap(function (item) {
            return _this.masterOnRequest(item.id, item.request, item.options);
        }), operators_1.map(function (response) { return _this.queueOut.next(response); }), 
        // Throwing here would end subscription, catch and log.
        operators_1.catchError(function (error) {
            _this.log.error(error);
            return rxjs_1.of(undefined);
        }))
            .subscribe();
    };
    /** Complete internal observable state, must be called by subclass on destruction. */
    Master.prototype.masterDestroy = function () {
        this.buffer = Buffer.allocUnsafe(0);
        this.receive.complete();
        this.transmit.complete();
        this.queueIn.complete();
        this.queueOut.complete();
    };
    /**
     * Receive data into internal buffer and try to parse responses.
     * Must be called by subclass after receiving data via connection for internal buffer management.
     * If a response is succesfully parsed then `masterOnResponse` shall be called with the result.
     */
    Master.prototype.masterOnData = function (data) {
        this.log.bytesReceived(data.length);
        this.buffer = Buffer.concat([this.buffer, data]);
        var parsedLength = this.masterParseResponse(this.buffer);
        this.buffer = this.buffer.slice(parsedLength);
    };
    /**
     * Response or exception parsed from incoming data by subclass.
     * Must be called by subclass after a response is successfully parsed from data.
     */
    Master.prototype.masterOnResponse = function (response) {
        if (pdu.isResponse(response)) {
            this.log.response(response);
            this.receive.next(response);
        }
        else if (pdu.isException(response)) {
            this.log.exception(response);
            this.receive.next(response);
        }
        else {
            this.log.error(response);
        }
    };
    // End implementation methods.
    // ---------------------------
    /** Validate number of retries (0+). */
    Master.prototype.isRetry = function (value) {
        value = value != null ? value : this.retry;
        return validate_1.isInteger(String(value), { min: 0 });
    };
    /** Validate timeout in milliseconds (50+). */
    Master.prototype.isTimeout = function (value) {
        value = value != null ? value : this.timeout;
        return validate_1.isInteger(String(value), { min: 50 });
    };
    /** Get retryWhen callback. */
    Master.prototype.isRetryWhen = function (callback) {
        return callback || this.retryWhen;
    };
    /** Master queue handler for requests. */
    Master.prototype.masterOnQueue = function (request, options) {
        // Generate ID using counter and emit on input queue.
        var id = ++this.queueCounter;
        this.queueIn.next({ id: id, request: request, options: options });
        // Filter queue output to get response/error for this request.
        return this.queueOut.pipe(operators_1.filter(function (out) { return out.id === id; }), operators_1.take(1), operators_1.mergeMap(function (out) {
            if (out.response != null) {
                return rxjs_1.of(out.response);
            }
            return rxjs_1.throwError(out.error);
        }));
    };
    /** Master handler for requests, sends request, waits for and matches responses. */
    Master.prototype.masterOnRequest = function (id, request, options) {
        var _this = this;
        var retry = this.isRetry(options.retry);
        var timeout = this.isTimeout(options.timeout);
        var retryWhen = this.isRetryWhen(options.retryWhen);
        // Transmit request via subject.
        this.log.request(request, 0);
        this.log.bytesTransmitted(request.buffer.length);
        this.transmit.next(request);
        // Wait for responses on receive subject.
        return this.receive.pipe(
        // Match responses to requests using subclass method.
        operators_1.filter(function (response) { return _this.masterMatchResponse(request, response); }), 
        // Timeout based on options.
        operators_1.timeout(timeout), 
        // Handle retries using retryWhen callback.
        operators_1.retryWhen(function (errors) {
            return errors.pipe(operators_1.scan(function (errorCount, error) {
                // Throws error if retry not required.
                errorCount += 1;
                retryWhen(_this, retry, errorCount, error, request);
                // Retransmit request and increment error counter.
                _this.log.request(request, errorCount);
                _this.log.bytesTransmitted(request.buffer.length);
                _this.transmit.next(request);
                return errorCount;
            }, 0));
        }), operators_1.take(1), operators_1.map(function (response) {
            // Return response or error if exception.
            if (pdu.isResponse(response)) {
                return { id: id, response: response };
            }
            else if (pdu.isException(response)) {
                return { id: id, error: response };
            }
            else {
                throw response;
            }
        }), operators_1.catchError(function (error) {
            return rxjs_1.of({ id: id, error: error });
        }));
    };
    return Master;
}());
exports.Master = Master;
//# sourceMappingURL=master.js.map