"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable:no-bitwise
var validate_1 = require("container.ts/lib/validate");
/** Modbus function codes. */
var EFunctionCode;
(function (EFunctionCode) {
    EFunctionCode[EFunctionCode["ReadCoils"] = 1] = "ReadCoils";
    EFunctionCode[EFunctionCode["ReadDiscreteInputs"] = 2] = "ReadDiscreteInputs";
    EFunctionCode[EFunctionCode["ReadHoldingRegisters"] = 3] = "ReadHoldingRegisters";
    EFunctionCode[EFunctionCode["ReadInputRegisters"] = 4] = "ReadInputRegisters";
    EFunctionCode[EFunctionCode["WriteSingleCoil"] = 5] = "WriteSingleCoil";
    EFunctionCode[EFunctionCode["WriteSingleRegister"] = 6] = "WriteSingleRegister";
    EFunctionCode[EFunctionCode["WriteMultipleCoils"] = 15] = "WriteMultipleCoils";
    EFunctionCode[EFunctionCode["WriteMultipleRegisters"] = 16] = "WriteMultipleRegisters";
    EFunctionCode[EFunctionCode["Mei"] = 43] = "Mei";
})(EFunctionCode = exports.EFunctionCode || (exports.EFunctionCode = {}));
/** Modbus exception codes. */
var EExceptionCode;
(function (EExceptionCode) {
    EExceptionCode[EExceptionCode["IllegalFunctionCode"] = 1] = "IllegalFunctionCode";
    EExceptionCode[EExceptionCode["IllegalDataAddress"] = 2] = "IllegalDataAddress";
    EExceptionCode[EExceptionCode["IllegalDataValue"] = 3] = "IllegalDataValue";
    EExceptionCode[EExceptionCode["ServerFailure"] = 4] = "ServerFailure";
    EExceptionCode[EExceptionCode["Acknowledge"] = 5] = "Acknowledge";
    EExceptionCode[EExceptionCode["ServerBusy"] = 6] = "ServerBusy";
})(EExceptionCode = exports.EExceptionCode || (exports.EExceptionCode = {}));
/** Modbus PDU request. */
var Request = /** @class */ (function () {
    function Request(functionCode, buffer) {
        this.functionCode = functionCode;
        this.buffer = buffer;
        this.isRequest = true;
    }
    return Request;
}());
exports.Request = Request;
/** Modbus PDU response. */
var Response = /** @class */ (function () {
    function Response(functionCode, data, buffer) {
        this.functionCode = functionCode;
        this.data = data;
        this.buffer = buffer;
        this.isResponse = true;
    }
    return Response;
}());
exports.Response = Response;
/** Modbus PDU exception. */
var Exception = /** @class */ (function () {
    function Exception(functionCode, exceptionFunctionCode, exceptionCode, buffer) {
        this.functionCode = functionCode;
        this.exceptionFunctionCode = exceptionFunctionCode;
        this.exceptionCode = exceptionCode;
        this.buffer = buffer;
        this.isException = true;
    }
    return Exception;
}());
exports.Exception = Exception;
/** Returns true if instance of Request. */
function isRequest(request) {
    var instanceOf = request instanceof Request;
    var hasProperty = !!request.isRequest;
    return instanceOf || hasProperty;
}
exports.isRequest = isRequest;
/** Returns true if instance of Response. */
function isResponse(response) {
    var instanceOf = response instanceof Response;
    var hasProperty = !!response.isResponse;
    return instanceOf || hasProperty;
}
exports.isResponse = isResponse;
/** Returns true if instance of Exception. */
function isException(exception) {
    var instanceOf = exception instanceof Exception;
    var hasProperty = !!exception.isException;
    return instanceOf || hasProperty;
}
exports.isException = isException;
/** Throw an error if value is not a valid address. */
function isAddress(value) {
    validate_1.isInteger(String(value), { min: 0x0, max: 0xffff });
}
exports.isAddress = isAddress;
/** Throw an error if value is not a valid register. */
function isRegister(value) {
    validate_1.isInteger(String(value), { min: 0x0, max: 0xffff });
}
exports.isRegister = isRegister;
/** Throw an error if value is not a valid quantity of bits. */
function isQuantityOfBits(value, maximum) {
    if (maximum === void 0) { maximum = 0x7d0; }
    validate_1.isInteger(String(value), { min: 0x1, max: maximum });
}
exports.isQuantityOfBits = isQuantityOfBits;
/** Throw an error if value is not a valid quantity of bits. */
function isQuantityOfRegisters(value, maximum) {
    if (maximum === void 0) { maximum = 0x7d; }
    validate_1.isInteger(String(value), { min: 0x1, max: maximum });
}
exports.isQuantityOfRegisters = isQuantityOfRegisters;
/** Convert an array of boolean bits to an array of byte values. */
function bitsToBytes(values) {
    var byteCount = Math.floor(values.length / 8);
    if (values.length % 8 !== 0) {
        byteCount += 1;
    }
    // Convert array of booleans to byte flag array.
    var byteValues = [];
    values.map(function (value, index) {
        var byteIndex = Math.floor(index / 8);
        var bitIndex = Math.floor(index % 8);
        var byteValue = byteValues[byteIndex] || 0;
        if (!!value) {
            byteValue |= 0x1 << bitIndex;
        }
        else {
            byteValue &= ~(0x1 << bitIndex);
        }
        byteValues[byteIndex] = byteValue;
    });
    return [byteCount, byteValues];
}
exports.bitsToBytes = bitsToBytes;
/** Convert an array of byte values to an array of boolean bits. */
function bytesToBits(quantity, buffer) {
    var byteCount = Math.floor(quantity / 8);
    if (quantity % 8 !== 0) {
        byteCount += 1;
    }
    // Convert byte flag array to array of booleans.
    var bitValues = [];
    for (var i = 0; i < quantity; i++) {
        var byteIndex = Math.floor(i / 8);
        var bitIndex = Math.floor(i % 8);
        var byteValue = buffer.readUInt8(byteIndex);
        bitValues.push(!!(byteValue & (0x1 << bitIndex)));
    }
    return [byteCount, bitValues];
}
exports.bytesToBits = bytesToBits;
//# sourceMappingURL=pdu.js.map