/// <reference types="node" />
import * as pdu from "./pdu";
/** Modbus PDU slave response or exception. */
export declare type ISlaveResponse = pdu.Response | pdu.Exception;
/** Modbus abstract PDU slave. */
export declare abstract class Slave {
    abstract readCoils(startingAddress: number, quantityOfCoils: number): boolean[];
    abstract readDiscreteInputs(startingAddress: number, quantityOfInputs: number): boolean[];
    abstract readHoldingRegisters(startingAddress: number, quantityOfRegisters: number): number[];
    abstract readInputRegisters(startingAddress: number, quantityOfRegisters: number): number[];
    abstract writeSingleCoil(outputAddress: number, outputValue: boolean): boolean;
    abstract writeSingleRegister(registerAddress: number, registerValue: number): number;
    abstract writeMultipleCoils(startingAddress: number, outputValues: boolean[]): number;
    abstract writeMultipleRegisters(startingAddress: number, registerValues: number[]): number;
    /**
     * Parse request buffer into PDU response or exception.
     * @param buffer Request buffer.
     */
    onRequest(buffer: Buffer): ISlaveResponse;
    protected onReadBitsRequest(functionCode: number, request: Buffer, isCoils: boolean): ISlaveResponse;
    protected onReadRegistersRequest(functionCode: number, request: Buffer, isHolding: boolean): ISlaveResponse;
    protected onWriteSingleBitRequest(functionCode: number, request: Buffer): ISlaveResponse;
    protected onWriteSingleRegisterRequest(functionCode: number, request: Buffer): ISlaveResponse;
    protected onWriteMultipleBitsRequest(functionCode: number, request: Buffer): ISlaveResponse;
    protected onWriteMultipleRegistersRequest(functionCode: number, request: Buffer): ISlaveResponse;
}
