/// <reference types="node" />
import { Observable } from "rxjs";
import * as adu from "../adu";
import * as pdu from "../pdu";
import * as rtu from "./rtu";
/** RTU master request options. */
export declare type IMasterRequestOptions = adu.IMasterRequestOptions<rtu.Request, rtu.Response, rtu.Exception>;
/** Generic serial port interface. */
export interface IMasterSerialPort {
    readonly isOpen: boolean;
    open(success?: () => void): void;
    close(): void;
    write(buffer: Buffer): void;
    on(event: "error", callback?: (error?: any) => void): this;
    on(event: "open" | "close", callback?: () => void): this;
    on(event: "data", callback?: (buffer: Buffer) => void): this;
}
/** Modbus RTU master options */
export interface IMasterOptions extends IMasterRequestOptions {
    slaveAddress?: number;
    inactivityTimeout?: number;
}
/** Modbus RTU master. */
export declare class Master extends adu.Master<rtu.Request, rtu.Response, rtu.Exception> {
    protected readonly port: IMasterSerialPort;
    readonly slaveAddress: number;
    readonly inactivityTimeout: number;
    constructor(port: IMasterSerialPort, options: IMasterOptions);
    open(options?: IMasterRequestOptions): Observable<void>;
    close(): void;
    destroy(): void;
    /** Write request to master port. */
    protected writePort(request: rtu.Request): void;
    /** Generate CRC for request buffer. */
    protected generateCrc(request: Buffer): number;
    protected onParseResponse(slaveAddress: number, pduBuffer: Buffer, aduBuffer: Buffer): rtu.Response | rtu.Exception | null;
    /** Construct and prepend RTU header, append CRC to PDU request buffer. */
    protected masterSetupRequest(functionCode: pdu.EFunctionCode, request: Buffer): rtu.Request;
    /** Incoming responses matched using expected size. */
    protected masterMatchResponse(request: rtu.Request, response: rtu.Response | rtu.Exception): boolean;
    protected masterParseResponse(data: Buffer): number;
}
