"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
// tslint:disable:no-bitwise prefer-for-of
var validate_1 = require("container.ts/lib/validate");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var adu = require("../adu");
var pdu = require("../pdu");
var rtu = require("./rtu");
/** Modbus RTU master. */
var Master = /** @class */ (function (_super) {
    __extends(Master, _super);
    function Master(port, options) {
        var _this = _super.call(this, options) || this;
        _this.port = port;
        _this.slaveAddress = options.slaveAddress != null ? validate_1.isInteger("" + options.slaveAddress, { min: 0, max: 0xff }) : 1;
        _this.inactivityTimeout = _this.isTimeout(options.inactivityTimeout);
        return _this;
    }
    Master.prototype.open = function (options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        // TODO(M): Timeout/retry support.
        return new rxjs_1.Observable(function (subscriber) {
            // Ensure master in known state.
            _this.masterReset();
            // (Re)open serial port
            if (!_this.port.isOpen) {
                _this.port.open();
            }
            // Add error listener.
            _this.port.on("error", function (error) {
                subscriber.error(new adu.MasterError(error.code, error));
            });
            // If port closes, call close and complete observable.
            var portClose = rxjs_1.fromEvent(_this.port, "close").pipe(operators_1.take(1));
            portClose.subscribe(function () {
                _this.close();
                subscriber.complete();
            });
            // If port opens, call next.
            rxjs_1.fromEvent(_this.port, "open")
                .pipe(operators_1.take(1))
                .subscribe(function () {
                subscriber.next();
            });
            // Port data event receives data into internal buffer and processes responses.
            rxjs_1.fromEvent(_this.port, "data")
                .pipe(operators_1.takeUntil(portClose))
                .subscribe(function (buffer) { return _this.masterOnData(buffer); });
            // Requests transmitted via port.
            _this.transmit.pipe(operators_1.takeUntil(portClose)).subscribe(function (request) { return _this.writePort(request); });
            // If no activity occurs on port for timeout duration, master is closed.
            rxjs_1.merge(_this.transmit, _this.receive)
                .pipe(operators_1.takeUntil(portClose), operators_1.timeout(_this.inactivityTimeout))
                .subscribe({
                error: function (error) {
                    _this.close();
                    subscriber.error(new adu.MasterError(error.code, error));
                }
            });
        });
    };
    Master.prototype.close = function () {
        if (this.port.isOpen) {
            this.port.close();
        }
        this.masterReset();
    };
    Master.prototype.destroy = function () {
        if (this.port.isOpen) {
            this.port.close();
        }
        this.masterDestroy();
    };
    /** Write request to master port. */
    Master.prototype.writePort = function (request) {
        if (this.port != null) {
            this.port.write(request.buffer);
        }
    };
    /** Generate CRC for request buffer. */
    Master.prototype.generateCrc = function (request) {
        var crc = 0xffff;
        for (var i = 0; i < request.length; i++) {
            crc = crc ^ request[i];
            for (var j = 0; j < 8; j++) {
                var odd = crc & 0x0001;
                crc = crc >> 1;
                if (!!odd) {
                    crc = crc ^ 0xa001;
                }
            }
        }
        return crc;
    };
    Master.prototype.onParseResponse = function (slaveAddress, pduBuffer, aduBuffer) {
        var pduResponse = pdu.Master.onResponse(pduBuffer);
        var response = null;
        if (pdu.isResponse(pduResponse)) {
            response = new rtu.Response(slaveAddress, pduResponse.functionCode, pduResponse.data, aduBuffer);
        }
        else if (pdu.isException(pduResponse)) {
            response = new rtu.Exception(slaveAddress, pduResponse.functionCode, pduResponse.exceptionFunctionCode, pduResponse.exceptionCode, aduBuffer);
        }
        return response;
    };
    /** Construct and prepend RTU header, append CRC to PDU request buffer. */
    Master.prototype.masterSetupRequest = function (functionCode, request) {
        var buffer = Buffer.concat([Buffer.allocUnsafe(1), request, Buffer.allocUnsafe(2)]);
        buffer.writeUInt8(this.slaveAddress, 0);
        var crc = this.generateCrc(buffer.slice(0, -2));
        buffer.writeUInt16LE(crc, request.length + 1);
        return new rtu.Request(this.slaveAddress, functionCode, buffer);
    };
    /** Incoming responses matched using expected size. */
    Master.prototype.masterMatchResponse = function (request, response) {
        return true;
    };
    Master.prototype.masterParseResponse = function (data) {
        // TODO(M): Improve parsing based on expected response.
        var slaveAddress = data.readUInt8(0);
        var pduBuffer = data.slice(1);
        var response = this.onParseResponse(slaveAddress, pduBuffer, data);
        if (response != null) {
            this.masterOnResponse(response);
        }
        // Return length of parsed data.
        return data.length;
    };
    return Master;
}(adu.Master));
exports.Master = Master;
//# sourceMappingURL=master.js.map