"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var validate_1 = require("container.ts/lib/validate");
var net_1 = require("net");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var adu = require("../adu");
var pdu = require("../pdu");
var tcp = require("../tcp");
/** Modbus TCP client log names. */
var EMasterLog;
(function (EMasterLog) {
    EMasterLog["Connecting"] = "ModbusTcpClientConnecting";
    EMasterLog["Connected"] = "ModbusTcpClientConnected";
    EMasterLog["Disconnected"] = "ModbusTcpClientDisconnected";
    EMasterLog["Error"] = "ModbusTcpClientError";
})(EMasterLog = exports.EMasterLog || (exports.EMasterLog = {}));
var ClientLog = /** @class */ (function (_super) {
    __extends(ClientLog, _super);
    function ClientLog() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    ClientLog.prototype.connecting = function (host, port, unitId) {
        this.debug(EMasterLog.Connecting, host, port, unitId);
    };
    ClientLog.prototype.connected = function (host, port, unitId) {
        this.debug(EMasterLog.Connected, host, port, unitId);
    };
    ClientLog.prototype.disconnected = function (host, port, unitId) {
        this.debug(EMasterLog.Disconnected, host, port, unitId);
    };
    ClientLog.prototype.error = function (error) {
        this.debug(EMasterLog.Error, error);
    };
    ClientLog.prototype.packetsTransmitted = function (value) { };
    ClientLog.prototype.packetsReceived = function (value) { };
    return ClientLog;
}(adu.MasterLog));
exports.ClientLog = ClientLog;
/** Modbus TCP client. */
var Client = /** @class */ (function (_super) {
    __extends(Client, _super);
    /**
     * Create TCP client instance.
     * @param options Client and default request options.
     */
    function Client(options, logConstructor) {
        if (logConstructor === void 0) { logConstructor = ClientLog; }
        var _this = _super.call(this, options, logConstructor) || this;
        _this.socket = null;
        _this.isConnected = false;
        _this.transactionId = 0;
        _this.host = options.host != null ? validate_1.isString(options.host) : "localhost";
        _this.port = options.port != null ? validate_1.isPort(String(options.port)) : 502;
        _this.unitId = options.unitId != null ? validate_1.isInteger(String(options.unitId), { min: 0x1, max: 0xff }) : 1;
        _this.inactivityTimeout = _this.isTimeout(options.inactivityTimeout);
        return _this;
    }
    Object.defineProperty(Client.prototype, "connected", {
        /** Is client connected. */
        get: function () {
            return this.isConnected;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Client.prototype, "connectionOptions", {
        /** Node socket connection options. */
        get: function () {
            return { host: this.host, port: this.port };
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(Client.prototype, "nextTransactionId", {
        /** Get next transaction identifier. */
        get: function () {
            this.transactionId = (this.transactionId + 1) % 0xffff;
            return this.transactionId;
        },
        enumerable: true,
        configurable: true
    });
    /**
     * Connect the client to configured host:port.
     * Observable will emit next after connected, throw an error if
     * connection fails and complete when 'disconnect' is called.
     * Completes if no data transmitted or received after timeout.
     * @param options Request options.
     */
    Client.prototype.connect = function (options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        var retry = this.isRetry(options.retry);
        var retryWhen = this.isRetryWhen(options.retryWhen);
        var retrying = false;
        return new rxjs_1.Observable(function (subscriber) {
            // Ensure master in known state.
            _this.masterReset();
            _this.destroySocket();
            _this.log.connecting(_this.host, _this.port, _this.unitId);
            // (Re)create socket, add error listener.
            _this.socket = net_1.createConnection(_this.connectionOptions);
            _this.socket.on("error", function (error) {
                subscriber.error(new adu.MasterError(error.code, error));
            });
            // If socket closes, call disconnect and complete observable.
            var socketClose = rxjs_1.fromEvent(_this.socket, "close").pipe(operators_1.take(1));
            socketClose.subscribe(function (hadError) {
                _this.disconnect();
                if (!retrying) {
                    subscriber.complete();
                }
            });
            // If socket connects, call next.
            rxjs_1.fromEvent(_this.socket, "connect")
                .pipe(operators_1.take(1))
                .subscribe(function () {
                _this.log.connected(_this.host, _this.port, _this.unitId);
                _this.isConnected = true;
                subscriber.next();
            });
            // Socket data event receives data into internal buffer and processes responses.
            rxjs_1.fromEvent(_this.socket, "data")
                .pipe(operators_1.takeUntil(socketClose))
                .subscribe(function (buffer) { return _this.masterOnData(buffer); });
            // Requests transmitted via socket.
            _this.transmit.pipe(operators_1.takeUntil(socketClose)).subscribe(function (request) { return _this.writeSocket(request); });
            // If no activity occurs on socket for timeout duration, client is disconnected.
            rxjs_1.merge(_this.transmit, _this.receive)
                .pipe(operators_1.takeUntil(socketClose), operators_1.timeout(_this.inactivityTimeout))
                .subscribe({
                error: function (error) {
                    _this.disconnect();
                    subscriber.error(error);
                }
            });
        }).pipe(operators_1.retryWhen(function (errors) {
            return errors.pipe(operators_1.scan(function (errorCount, error) {
                errorCount += 1;
                retrying = false;
                // Throws error if retry not required.
                retryWhen(_this, retry, errorCount, error);
                retrying = true;
                return errorCount;
            }, 0), operators_1.delay(100));
        }));
    };
    /** If connected, disconnect the client from the configured host:port. */
    Client.prototype.disconnect = function () {
        this.log.disconnected(this.host, this.port, this.unitId);
        this.destroySocket();
        this.masterReset();
    };
    Client.prototype.destroy = function () {
        this.destroySocket();
        this.masterDestroy();
    };
    Client.prototype.destroySocket = function () {
        if (this.socket != null) {
            this.isConnected = false;
            this.socket.end();
            this.socket.destroy();
            this.socket = null;
        }
    };
    Client.prototype.writeSocket = function (request) {
        if (this.socket != null && this.isConnected) {
            this.socket.write(request.buffer);
            this.log.packetsTransmitted(1);
        }
        else {
            this.log.error(new adu.MasterError(adu.EMasterError.Write));
        }
    };
    /**
     * Parse buffer into response.
     * Inheritors may overwrite this function to implement their own response handler.
     */
    Client.prototype.onParseResponse = function (transactionId, unitId, pduBuffer, aduBuffer) {
        var pduResponse = pdu.Master.onResponse(pduBuffer);
        var response = null;
        if (pdu.isResponse(pduResponse)) {
            response = new tcp.Response(transactionId, unitId, pduResponse.functionCode, pduResponse.data, aduBuffer);
        }
        else if (pdu.isException(pduResponse)) {
            response = new tcp.Exception(transactionId, unitId, pduResponse.functionCode, pduResponse.exceptionFunctionCode, pduResponse.exceptionCode, aduBuffer);
        }
        return response;
    };
    /** Construct and prepend MBAP header to PDU request buffer. */
    Client.prototype.masterSetupRequest = function (functionCode, request) {
        var buffer = Buffer.concat([Buffer.allocUnsafe(7), request]);
        var transactionId = this.nextTransactionId;
        buffer.writeUInt16BE(transactionId, 0);
        buffer.writeUInt16BE(0, 2); // Protocol ID.
        buffer.writeUInt16BE(request.length + 1, 4);
        buffer.writeUInt8(this.unitId, 6);
        return new tcp.Request(transactionId, this.unitId, functionCode, buffer);
    };
    /** Match transaction and unit identifiers of incoming responses. */
    Client.prototype.masterMatchResponse = function (request, response) {
        var transactionIdMatch = response.transactionId === request.transactionId;
        var unitIdMatch = response.unitId === request.unitId;
        return transactionIdMatch && unitIdMatch;
    };
    Client.prototype.masterParseResponse = function (data) {
        // Check if buffer may contain MBAP header.
        if (data.length >= 7) {
            var header = data.slice(0, 7);
            var headerLength = header.readUInt16BE(4);
            var responseLength = 6 + headerLength;
            // If buffer contains complete response, extract it now.
            if (data.length >= responseLength) {
                this.log.packetsReceived(1);
                var aduBuffer = data.slice(0, responseLength);
                var transactionId = aduBuffer.readUInt16BE(0);
                var unitId = aduBuffer.readUInt8(6);
                var pduBuffer = aduBuffer.slice(7);
                var response = this.onParseResponse(transactionId, unitId, pduBuffer, aduBuffer);
                if (response != null) {
                    this.masterOnResponse(response);
                }
                // Return length of parsed data.
                return responseLength;
            }
        }
        return 0;
    };
    return Client;
}(adu.Master));
exports.Client = Client;
//# sourceMappingURL=client.js.map