/// <reference types="node" />
import * as net from "net";
import { Observable, Subject } from "rxjs";
import * as pdu from "../pdu";
import * as tcp from "./tcp";
/** Modbus TCP observable response or exception. */
export declare type IServerResponse = tcp.Response | tcp.Exception | null;
export interface IServerOptions {
    port?: number;
}
export declare abstract class Server extends pdu.Slave {
    /** Port the server will listen on. */
    readonly port: number;
    protected server: net.Server | null;
    constructor(options: IServerOptions);
    open(): Observable<void>;
    close(): void;
    protected onData(buffer: Buffer, data: Buffer, transmit: Subject<tcp.Response | tcp.Exception>): Buffer;
    protected onParseResponse(transactionId: number, unitId: number, pduBuffer: Buffer): IServerResponse;
    protected aduHeader(response: tcp.Response | tcp.Exception): Buffer;
    protected writeSocket(socket: net.Socket, packet: Buffer): void;
}
