"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = function (d, b) {
        extendStatics = Object.setPrototypeOf ||
            ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
            function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
        return extendStatics(d, b);
    };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var validate_1 = require("container.ts/lib/validate");
var net = require("net");
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var pdu = require("../pdu");
var tcp = require("./tcp");
var Server = /** @class */ (function (_super) {
    __extends(Server, _super);
    function Server(options) {
        var _this = _super.call(this) || this;
        _this.server = null;
        _this.port = options.port != null ? validate_1.isPort(String(options.port)) : 502;
        return _this;
    }
    Server.prototype.open = function () {
        var _this = this;
        // (Re)create server instance.
        this.server = net.createServer(function (socket) {
            var transmit = new rxjs_1.Subject();
            // Map socket close event to observable.
            // Close event will complete other socket observables.
            var socketClose = rxjs_1.fromEvent(socket, "close").pipe(operators_1.take(1));
            socketClose.subscribe(function () {
                transmit.complete();
            });
            // Receive data into buffer and process.
            var buffer = Buffer.allocUnsafe(0);
            var socketData = rxjs_1.fromEvent(socket, "data").pipe(operators_1.takeUntil(socketClose));
            socketData.subscribe(function (data) {
                buffer = _this.onData(buffer, data, transmit);
            });
            // Transmit responses by writing to socket.
            transmit.pipe(operators_1.takeUntil(socketClose)).subscribe(function (response) {
                var packet = _this.aduHeader(response);
                _this.writeSocket(socket, packet);
            });
        });
        // Server listen to port.
        var serverListen = rxjs_1.bindCallback(this.server.listen.bind(this.server, this.port));
        return serverListen().pipe(operators_1.map(function () { return undefined; }));
    };
    Server.prototype.close = function () {
        if (this.server != null) {
            this.server.close();
            this.server = null;
        }
    };
    Server.prototype.onData = function (buffer, data, transmit) {
        buffer = Buffer.concat([buffer, data]);
        // Check if buffer may container MBAP header.
        if (buffer.length >= 7) {
            var header = buffer.slice(0, 7);
            var headerLength = header.readUInt16BE(4);
            var requestLength = 6 + headerLength;
            // If buffer contains complete request, extract it now.
            if (buffer.length >= requestLength) {
                var aduBuffer = buffer.slice(0, requestLength);
                var transactionId = aduBuffer.readUInt16BE(0);
                var unitId = aduBuffer.readUInt8(6);
                var pduBuffer = aduBuffer.slice(7);
                // Parse PDU slice of buffer.
                // Inheritors may overwrite this function to implement their own handling.
                var response = this.onParseResponse(transactionId, unitId, pduBuffer);
                if (response != null) {
                    transmit.next(response);
                }
                // Return buffer with packet removed.
                return buffer.slice(requestLength);
            }
        }
        // Return concatenated buffers.
        return buffer;
    };
    Server.prototype.onParseResponse = function (transactionId, unitId, pduBuffer) {
        var pduResponse = this.onRequest(pduBuffer);
        var response = null;
        if (pdu.isResponse(pduResponse)) {
            response = new tcp.Response(transactionId, unitId, pduResponse.functionCode, pduResponse.data, pduResponse.buffer);
        }
        else if (pdu.isException(pduResponse)) {
            response = new tcp.Exception(transactionId, unitId, pduResponse.functionCode, pduResponse.exceptionFunctionCode, pduResponse.exceptionCode, pduResponse.buffer);
        }
        return response;
    };
    Server.prototype.aduHeader = function (response) {
        var buffer = Buffer.concat([Buffer.allocUnsafe(7), response.buffer]);
        buffer.writeUInt16BE(response.transactionId, 0);
        buffer.writeUInt16BE(0, 2);
        buffer.writeUInt16BE(response.buffer.length + 1, 4);
        buffer.writeUInt8(response.unitId, 6);
        return buffer;
    };
    Server.prototype.writeSocket = function (socket, packet) {
        socket.write(packet);
    };
    return Server;
}(pdu.Slave));
exports.Server = Server;
//# sourceMappingURL=server.js.map